
package org.mule.module.google.calendar.processors;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import org.mule.api.MessagingException;
import org.mule.api.MuleContext;
import org.mule.api.MuleEvent;
import org.mule.api.MuleException;
import org.mule.api.MuleMessage;
import org.mule.api.construct.FlowConstruct;
import org.mule.api.construct.FlowConstructAware;
import org.mule.api.context.MuleContextAware;
import org.mule.api.expression.ExpressionManager;
import org.mule.api.lifecycle.Disposable;
import org.mule.api.lifecycle.Initialisable;
import org.mule.api.lifecycle.InitialisationException;
import org.mule.api.lifecycle.Startable;
import org.mule.api.lifecycle.Stoppable;
import org.mule.api.processor.MessageProcessor;
import org.mule.api.registry.RegistrationException;
import org.mule.api.transformer.DataType;
import org.mule.api.transformer.Transformer;
import org.mule.api.transformer.TransformerException;
import org.mule.config.i18n.CoreMessages;
import org.mule.config.i18n.MessageFactory;
import org.mule.module.google.calendar.GoogleCalendarConnectorModule;
import org.mule.module.google.calendar.adapters.GoogleCalendarConnectorModuleOAuth2Adapter;
import org.mule.transformer.TransformerTemplate;
import org.mule.transformer.types.DataTypeFactory;
import org.mule.transport.NullPayload;
import org.mule.util.TemplateParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * GetEventsMessageProcessor invokes the {@link org.mule.module.google.calendar.GoogleCalendarConnectorModule#getEvents(java.lang.String, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer, java.lang.String, java.lang.String, java.lang.String, boolean, boolean, boolean, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String)} method in {@link GoogleCalendarConnectorModule }. For each argument there is a field in this processor to match it.  Before invoking the actual method the processor will evaluate and transform where possible to the expected argument type.
 * 
 */
public class GetEventsMessageProcessor
    implements FlowConstructAware, MuleContextAware, Disposable, Initialisable, Startable, Stoppable, MessageProcessor
{

    private Object accessToken;
    private String _accessTokenType;
    private Object calendarId;
    private String _calendarIdType;
    private Object icalUID;
    private String _icalUIDType;
    private Object maxAttendees;
    private Integer _maxAttendeesType;
    private Object maxResults;
    private Integer _maxResultsType;
    private Object orderBy;
    private String _orderByType;
    private Object pageToken;
    private String _pageTokenType;
    private Object query;
    private String _queryType;
    private Object showDeleted;
    private boolean _showDeletedType;
    private Object showHiddenInvitations;
    private boolean _showHiddenInvitationsType;
    private Object singleEvents;
    private boolean _singleEventsType;
    private Object timeMin;
    private String _timeMinType;
    private Object timeMax;
    private String _timeMaxType;
    private Object datetimeFormat;
    private String _datetimeFormatType;
    private Object timezone;
    private String _timezoneType;
    private Object lastUpdated;
    private String _lastUpdatedType;
    private static Logger logger = LoggerFactory.getLogger(GetEventsMessageProcessor.class);
    /**
     * Module object
     * 
     */
    private Object moduleObject;
    /**
     * Mule Context
     * 
     */
    private MuleContext muleContext;
    /**
     * Mule Expression Manager
     * 
     */
    private ExpressionManager expressionManager;
    /**
     * Mule Pattern Info
     * 
     */
    private TemplateParser.PatternInfo patternInfo;
    /**
     * Flow construct
     * 
     */
    private FlowConstruct flowConstruct;
    /**
     * Variable used to track how many retries we have attempted on this message processor
     * 
     */
    private AtomicInteger retryCount;
    /**
     * Maximum number of retries that can be attempted.
     * 
     */
    private int retryMax;

    /**
     * Obtains the expression manager from the Mule context and initialises the connector. If a target object  has not been set already it will search the Mule registry for a default one.
     * 
     * @throws InitialisationException
     */
    public void initialise()
        throws InitialisationException
    {
        retryCount = new AtomicInteger();
        expressionManager = muleContext.getExpressionManager();
        patternInfo = TemplateParser.createMuleStyleParser().getStyle();
        if (moduleObject == null) {
            try {
                moduleObject = muleContext.getRegistry().lookupObject(GoogleCalendarConnectorModuleOAuth2Adapter.class);
                if (moduleObject == null) {
                    throw new InitialisationException(MessageFactory.createStaticMessage("Cannot find object"), this);
                }
            } catch (RegistrationException e) {
                throw new InitialisationException(CoreMessages.initialisationFailure("org.mule.module.google.calendar.adapters.GoogleCalendarConnectorModuleOAuth2Adapter"), e, this);
            }
        }
        if (moduleObject instanceof String) {
            moduleObject = muleContext.getRegistry().lookupObject(((String) moduleObject));
            if (moduleObject == null) {
                throw new InitialisationException(MessageFactory.createStaticMessage("Cannot find object by config name"), this);
            }
        }
    }

    public void start()
        throws MuleException
    {
    }

    public void stop()
        throws MuleException
    {
    }

    public void dispose() {
    }

    /**
     * Set the Mule context
     * 
     * @param context Mule context to set
     */
    public void setMuleContext(MuleContext context) {
        this.muleContext = context;
    }

    /**
     * Sets flow construct
     * 
     * @param flowConstruct Flow construct to set
     */
    public void setFlowConstruct(FlowConstruct flowConstruct) {
        this.flowConstruct = flowConstruct;
    }

    /**
     * Sets the instance of the object under which the processor will execute
     * 
     * @param moduleObject Instace of the module
     */
    public void setModuleObject(Object moduleObject) {
        this.moduleObject = moduleObject;
    }

    /**
     * Sets retryMax
     * 
     * @param value Value to set
     */
    public void setRetryMax(int value) {
        this.retryMax = value;
    }

    /**
     * Sets calendarId
     * 
     * @param value Value to set
     */
    public void setCalendarId(Object value) {
        this.calendarId = value;
    }

    /**
     * Sets accessToken
     * 
     * @param value Value to set
     */
    public void setAccessToken(Object value) {
        this.accessToken = value;
    }

    /**
     * Sets orderBy
     * 
     * @param value Value to set
     */
    public void setOrderBy(Object value) {
        this.orderBy = value;
    }

    /**
     * Sets icalUID
     * 
     * @param value Value to set
     */
    public void setIcalUID(Object value) {
        this.icalUID = value;
    }

    /**
     * Sets query
     * 
     * @param value Value to set
     */
    public void setQuery(Object value) {
        this.query = value;
    }

    /**
     * Sets datetimeFormat
     * 
     * @param value Value to set
     */
    public void setDatetimeFormat(Object value) {
        this.datetimeFormat = value;
    }

    /**
     * Sets lastUpdated
     * 
     * @param value Value to set
     */
    public void setLastUpdated(Object value) {
        this.lastUpdated = value;
    }

    /**
     * Sets pageToken
     * 
     * @param value Value to set
     */
    public void setPageToken(Object value) {
        this.pageToken = value;
    }

    /**
     * Sets singleEvents
     * 
     * @param value Value to set
     */
    public void setSingleEvents(Object value) {
        this.singleEvents = value;
    }

    /**
     * Sets showHiddenInvitations
     * 
     * @param value Value to set
     */
    public void setShowHiddenInvitations(Object value) {
        this.showHiddenInvitations = value;
    }

    /**
     * Sets maxResults
     * 
     * @param value Value to set
     */
    public void setMaxResults(Object value) {
        this.maxResults = value;
    }

    /**
     * Sets maxAttendees
     * 
     * @param value Value to set
     */
    public void setMaxAttendees(Object value) {
        this.maxAttendees = value;
    }

    /**
     * Sets timezone
     * 
     * @param value Value to set
     */
    public void setTimezone(Object value) {
        this.timezone = value;
    }

    /**
     * Sets timeMin
     * 
     * @param value Value to set
     */
    public void setTimeMin(Object value) {
        this.timeMin = value;
    }

    /**
     * Sets showDeleted
     * 
     * @param value Value to set
     */
    public void setShowDeleted(Object value) {
        this.showDeleted = value;
    }

    /**
     * Sets timeMax
     * 
     * @param value Value to set
     */
    public void setTimeMax(Object value) {
        this.timeMax = value;
    }

    /**
     * Get all superclasses and interfaces recursively.
     * 
     * @param classes List of classes to which to add all found super classes and interfaces.
     * @param clazz   The class to start the search with.
     */
    private void computeClassHierarchy(Class clazz, List classes) {
        for (Class current = clazz; (current!= null); current = current.getSuperclass()) {
            if (classes.contains(current)) {
                return ;
            }
            classes.add(current);
            for (Class currentInterface: current.getInterfaces()) {
                computeClassHierarchy(currentInterface, classes);
            }
        }
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link List }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isListClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(List.class);
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link Map }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isMapClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(Map.class);
    }

    private boolean isList(Type type) {
        if ((type instanceof Class)&&isListClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isList(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isList(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isMap(Type type) {
        if ((type instanceof Class)&&isMapClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isMap(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isMap(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isAssignableFrom(Type expectedType, Class clazz) {
        if (expectedType instanceof Class) {
            if (((Class) expectedType).isPrimitive()) {
                if (((Class) expectedType).getName().equals("boolean")&&(clazz == Boolean.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("byte")&&(clazz == Byte.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("short")&&(clazz == Short.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("char")&&(clazz == Character.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("int")&&(clazz == Integer.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("float")&&(clazz == Float.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("long")&&(clazz == Long.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("double")&&(clazz == Double.class)) {
                    return true;
                }
                return false;
            } else {
                return ((Class) expectedType).isAssignableFrom(clazz);
            }
        }
        if (expectedType instanceof ParameterizedType) {
            return isAssignableFrom(((ParameterizedType) expectedType).getRawType(), clazz);
        }
        if (expectedType instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) expectedType).getUpperBounds();
            if (upperBounds.length!= 0) {
                return isAssignableFrom(upperBounds[ 0 ], clazz);
            }
        }
        return false;
    }

    private Object evaluate(MuleMessage muleMessage, Object source) {
        if (source instanceof String) {
            String stringSource = ((String) source);
            if (stringSource.startsWith(patternInfo.getPrefix())&&stringSource.endsWith(patternInfo.getSuffix())) {
                return expressionManager.evaluate(stringSource, muleMessage);
            } else {
                return expressionManager.parse(stringSource, muleMessage);
            }
        }
        return source;
    }

    private Object evaluateAndTransform(MuleMessage muleMessage, Type expectedType, String expectedMimeType, Object source)
        throws TransformerException
    {
        if (source == null) {
            return source;
        }
        Object target = null;
        if (isList(source.getClass())) {
            if (isList(expectedType)) {
                List newList = new ArrayList();
                Type valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                ListIterator iterator = ((List) source).listIterator();
                while (iterator.hasNext()) {
                    Object subTarget = iterator.next();
                    newList.add(evaluateAndTransform(muleMessage, valueType, expectedMimeType, subTarget));
                }
                target = newList;
            } else {
                target = source;
            }
        } else {
            if (isMap(source.getClass())) {
                if (isMap(expectedType)) {
                    Type keyType = Object.class;
                    Type valueType = Object.class;
                    if (expectedType instanceof ParameterizedType) {
                        keyType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                        valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 1 ];
                    }
                    Map map = ((Map) source);
                    Map newMap = new HashMap();
                    for (Object entryObj: map.entrySet()) {
                        {
                            Map.Entry entry = ((Map.Entry) entryObj);
                            Object newKey = evaluateAndTransform(muleMessage, keyType, expectedMimeType, entry.getKey());
                            Object newValue = evaluateAndTransform(muleMessage, valueType, expectedMimeType, entry.getValue());
                            newMap.put(newKey, newValue);
                        }
                    }
                    target = newMap;
                } else {
                    target = source;
                }
            } else {
                target = evaluate(muleMessage, source);
            }
        }
        if ((target!= null)&&(!isAssignableFrom(expectedType, target.getClass()))) {
            DataType sourceDataType = DataTypeFactory.create(target.getClass());
            DataType targetDataType = null;
            if (expectedMimeType!= null) {
                targetDataType = DataTypeFactory.create(((Class) expectedType), expectedMimeType);
            } else {
                targetDataType = DataTypeFactory.create(((Class) expectedType));
            }
            Transformer t = muleContext.getRegistry().lookupTransformer(sourceDataType, targetDataType);
            return t.transform(target);
        } else {
            return target;
        }
    }

    /**
     * Invokes the MessageProcessor.
     * 
     * @param event MuleEvent to be processed
     * @throws MuleException
     */
    public MuleEvent process(MuleEvent event)
        throws MuleException
    {
        MuleMessage _muleMessage = event.getMessage();
        GoogleCalendarConnectorModuleOAuth2Adapter _castedModuleObject = null;
        if (moduleObject instanceof String) {
            _castedModuleObject = ((GoogleCalendarConnectorModuleOAuth2Adapter) muleContext.getRegistry().lookupObject(((String) moduleObject)));
            if (_castedModuleObject == null) {
                throw new MessagingException(CoreMessages.failedToCreate("getEvents"), event, new RuntimeException("Cannot find the configuration specified by the config-ref attribute."));
            }
        } else {
            _castedModuleObject = ((GoogleCalendarConnectorModuleOAuth2Adapter) moduleObject);
        }
        try {
            String _transformedCalendarId = ((String) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_calendarIdType").getGenericType(), null, calendarId));
            String _transformedIcalUID = ((String) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_icalUIDType").getGenericType(), null, icalUID));
            Integer _transformedMaxAttendees = ((Integer) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_maxAttendeesType").getGenericType(), null, maxAttendees));
            Integer _transformedMaxResults = ((Integer) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_maxResultsType").getGenericType(), null, maxResults));
            String _transformedOrderBy = ((String) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_orderByType").getGenericType(), null, orderBy));
            String _transformedPageToken = ((String) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_pageTokenType").getGenericType(), null, pageToken));
            String _transformedQuery = ((String) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_queryType").getGenericType(), null, query));
            Boolean _transformedShowDeleted = ((Boolean) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_showDeletedType").getGenericType(), null, showDeleted));
            Boolean _transformedShowHiddenInvitations = ((Boolean) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_showHiddenInvitationsType").getGenericType(), null, showHiddenInvitations));
            Boolean _transformedSingleEvents = ((Boolean) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_singleEventsType").getGenericType(), null, singleEvents));
            String _transformedTimeMin = ((String) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_timeMinType").getGenericType(), null, timeMin));
            String _transformedTimeMax = ((String) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_timeMaxType").getGenericType(), null, timeMax));
            String _transformedDatetimeFormat = ((String) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_datetimeFormatType").getGenericType(), null, datetimeFormat));
            String _transformedTimezone = ((String) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_timezoneType").getGenericType(), null, timezone));
            String _transformedLastUpdated = ((String) evaluateAndTransform(_muleMessage, GetEventsMessageProcessor.class.getDeclaredField("_lastUpdatedType").getGenericType(), null, lastUpdated));
            retryCount.getAndIncrement();
            Object resultPayload;
            resultPayload = _castedModuleObject.getEvents(_transformedCalendarId, _transformedIcalUID, _transformedMaxAttendees, _transformedMaxResults, _transformedOrderBy, _transformedPageToken, _transformedQuery, _transformedShowDeleted, _transformedShowHiddenInvitations, _transformedSingleEvents, _transformedTimeMin, _transformedTimeMax, _transformedDatetimeFormat, _transformedTimezone, _transformedLastUpdated);
            TransformerTemplate.OverwitePayloadCallback overwritePayloadCallback = null;
            if (resultPayload == null) {
                overwritePayloadCallback = new TransformerTemplate.OverwitePayloadCallback(NullPayload.getInstance());
            } else {
                overwritePayloadCallback = new TransformerTemplate.OverwitePayloadCallback(resultPayload);
            }
            List<Transformer> transformerList;
            transformerList = new ArrayList<Transformer>();
            transformerList.add(new TransformerTemplate(overwritePayloadCallback));
            event.getMessage().applyTransformers(event, transformerList);
            retryCount.set(0);
            return event;
        } catch (Exception e) {
            throw new MessagingException(CoreMessages.failedToInvoke("getEvents"), event, e);
        }
    }

}
